"""
The CLI module. Contains every command used by ``ezvi``.
No function other than the command line options should be
defined in this module.
"""
import click
from ezvi import funcmodule


@click.group()
def app():
    """A tool to automate typing in the Vi editor"""


@click.command()
@click.argument(
    "infile",
    type=click.File("r"),
)
@click.option(
    "-w",
    "--writefile",
    type=str,
    help="""\
    To save the newly created file. 
    Use `ezvi text -w [NEW_PATH] [PATH_TO_EXISTING_FILE]`.
    """,
)
def text(infile, writefile):
    """
    To re-type an already pre-typed file. `ezvi` will just rewrite the
    file as-is character by character.
    """
    if not writefile:
        writing = funcmodule.file_parser(infile)
    else:
        writing = funcmodule.file_parser(infile, name=writefile)

    funcmodule.ez_spawn(("vi",), writing)


@click.command()
@click.argument(
    "config",
    type=click.File("r"),
)
def yaml(config):
    """To use a YAML config file as input."""

    parsed = funcmodule.yaml_parser(config)
    writing = []
    for item in parsed:
        for key in item:
            writing.append(item[key])

    funcmodule.ez_spawn(("vi",), writing)


@click.command()
@click.argument(
    "infile",
    type=click.File("r"),
)
@click.option(
    "-s",
    "--savepath",
    type=str,
    help="""\
        To save the config file.
        This is the path towards where the file will be saved.
        """,
)
def create_config(infile, savepath):
    """To generate a config file."""
    if savepath:
        funcmodule.new_conf(infile, savepath)
    else:
        for line in infile:
            line = line.strip()
            if not line:
                to_echo = "- new_line: "
            else:
                to_echo = "- write_line: " + line
            click.echo(to_echo)


app.add_command(create_config)
app.add_command(yaml)
app.add_command(text)


def main():
    app()



"""
``tools.py`` contains functions used in the funcmodule to interact with
the ``vi`` editor. Most functions can be mapped to existing ``vi``
commands.

This module is also where the API is documented. Each funcion's docstring
contains documentation on how to use the function with the API and with
the command line interface.
"""
from typing import List, Dict, Callable, Any

ESCAPE: str = chr(27)

#######################################################################
#                       Character Encoding                            #
#######################################################################


def ez_encode_str(to_encode: str) -> list:
    """Encodes a `str` per character and puts them into a list.

    :type to_encode: str
    :param to_encode: The string that has to be encoded.

    :rtype: list
    :return: A list of encoded chars. Encodes in UTF-8
    """
    to_return: List[bytes] = []
    for char in list(to_encode):
        if not isinstance(char, bytes):
            try:
                to_return.append(char.encode("utf-8"))
            except AttributeError:
                raise Exception("`to_encode` must be of type `str`")
        else:
            # This is a problem as they could be encoded differently.
            to_return.append(char)
    return to_return


#######################################################################
#                            Vi tools                                 #
#######################################################################


# Writing
def write_chars(to_write: str) -> list:
    """To type ``to_write`` to the file.

    ``write_chars`` will type the passed string after the cursor position.
    From Vi’s command mode, it types ``a`` to insert after and then types the
    string.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_chars: "snek"

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_chars("snek")

    :type to_write: str
    :param to_write: The characters to write.

    :rtype: str
    :return: A string that corresponds to the keys that would be
    pressed to insert text after the current cursor position.
    """

    to_write = "a" + to_write + ESCAPE

    return to_write


def write_line(to_write: str) -> str:
    """To type `to_write` and create a new line.

    Starts typing after the current cursor position by pressing
    ``a`` from the command mode. ``to_write`` is then typed and
    a new line is created.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_line: "Python is fun."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_line("Python is fun.")

    :type to_write: str
    :param to_write: The characters to write.

    :rtype: str
    :return: A string that contains the keystrokes that would be used
    to write a whole line in ``Vi``.
    """

    to_write = "a" + to_write + "\n" + ESCAPE

    return to_write


def new_line(amount=1):
    """Creates an ``amount`` of new lines.

    ``new_line`` inserts a certain number of new lines to the file.
    From Vi’s command mode, ``ezvi`` first presses ``o``. This ensures
    that the current line won’t be split even if the cursor is not
    at the end of the line.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - new_line: 3


    `Using the API:`

    .. code-block:: python

      ezvi.tools.new_line(3)

    :type amount: int
    :param amount: The number of new lines to create.

    :rtype: str
    :return: A string that contains the characters that would be used
    to create a certain amount of new lines in ``Vi``.
    """
    # TODO: use ``isinstance``
    if type(amount) != int:
        try:
            amount = int(amount)
        except TypeError:
            amount = 1

    to_write = "o" + "\n" * (amount - 1) + ESCAPE

    return to_write


def new_line_over():
    """Creates a new line over the cursor.

    The cursor is also moved to the beginning of the new line. It is
    not possible to create more than one new line over the cursor
    at a time for now.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - new_line_over:


    `Using the API:`

    .. code-block:: python

      ezvi.tools.new_line_over()

    :rtype: str
    :return: Characters that would be used in ``Vi`` to add a new line
    over the cursor.
    """

    to_write = "O" + ESCAPE

    return to_write


def write_after_word(to_write):
    """To write ``to_write`` after the current word.

    This function uses ``e`` from the command mode to go to the end
    of the word. ``to_write`` is then written after the end of the
    word using the ``a`` command. **This function does not add a
    space to the beginning of ``to_write``.**

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_after_word: " General Kenobi."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_after_word(" General Kenobi.")

    :type to_write: str
    :param to_write: What to write after the word.

    :rtype: str
    :return: Characters that would be used to type contents after the
    current word.
    """

    prepend = "e" + "a"
    append = ESCAPE
    to_write = prepend + to_write + append

    return to_write


def write_after_line(to_write):
    """To write ``to_write`` after the current line.

    This function uses ``$`` from the command mode to go to the
    end of the line. ``to_write`` is then written after the cursor
    position using the ``a`` command. **This function does not add a
    space to the beginning of ``to_write``.**

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_after_line: " General Kenobi."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_after_line(" General Kenobi.")

    :type to_write: str
    :param to_write: What to write after the line.

    :rtype: str
    :return: Characters that would be used to append content to a line.
    """

    prepend = "$" + "a"
    append = ESCAPE
    to_write = prepend + to_write + append

    return to_write


def write_after_char(to_write):
    """To write ``to_write`` after the cursor's position (current char).

    ``to_write`` is written after the cursor
    position using the ``a`` command.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_after_char: "Greetings!"

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_after_char("Greetings!")

    :type to_write: str
    :param to_write: What to write after the cursor.

    :rtype: str
    :return: Characters that would be used to add contents after the 
    current cursor position.
    """

    prepend = "a"
    append = ESCAPE
    to_write = prepend + to_write + append

    return to_write


def write_before_word(to_write):
    """To write ``to_write`` before the current word.

    ``write_before_word`` uses ``b`` from the command mode to move
    the cursor to the beginning of the current word. ``to_write`` is
    then written before the cursor's position using the ``i``
    command.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_before_word: "Hello there."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_before_word("Hello there.")

    :type to_write: str
    :param to_write: What to write before the current word.

    :rtype: str
    :return: Keystrokes that can be used to prepend a word with new contents.
    """

    prepend = "b" + "i"  # TODO: Replace "b" by something that works.
    append = ESCAPE
    to_write = prepend + to_write + append

    return to_write


def write_before_line(to_write):
    """To write ``to_write`` at the beginning of the line.

    ``write_before_line`` uses ``0`` from the command mode to move
    the cursor to the beginning of the current line. ``to_write`` is
    then written before the cursor's position using the ``i``
    command.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_before_line: "Hello there."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_before_line("Hello there.")

    :type to_write: str
    :param to_write: What to write before the current line.

    :rtype: str
    :return: Characters that can be used to prepend content to a line.
    """

    prepend = "0" + "i"
    append = ESCAPE
    to_write = prepend + to_write + append

    return to_write


def write_before_char(to_write):
    """To write ``to_write`` at the beginning of the line.

    ``write_before_line`` uses ``0`` from the command mode to move
    the cursor to the beginning of the current line. ``to_write`` is
    then written before the cursor's position using the ``i``
    command.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_before_line: "Hello there."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_before_line("Hello there.")

    :type to_write: str
    :param to_write: What to write before the current line.

    :rtype: str
    :return: Keystrokes that can be used to add content before the
    current cursor position.
    """

    prepend = "i"
    append = ESCAPE
    to_write = prepend + to_write + append

    return to_write


# Movement


def goto_line(line_num):
    """To go to a certain line.

    This function uses the ``G`` command to move the cursor to
    the beginning of a certain line.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - goto_line: 5

    `Using the API:`

    .. code-block:: python

      ezvi.tools.goto_line(5)

    :type line_num: int
    :param line_num: The number of the line to move the cursor to.

    :rtype: str
    :return: A line number appended with ``G``.
    """

    to_write = str(line_num) + "G"

    return to_write


def goto_column(column_num):
    """To go to a certain column.

    This function uses the ``l`` command to move the cursor to
    a certain column on the **current line**.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - goto_column: 5

    `Using the API:`

    .. code-block:: python

      ezvi.tools.goto_column(5)

    :type column_num: int
    :param column_num: The number of the column to move the cursor to.

    :rtype: str
    :return: ``0`` plus a column position appended with ``l``.
    """

    # This would be much cleaner if I could get the cursor's position.
    to_write = "0" + str(column_num - 1) + "l"

    return to_write


# Replace functions


def replace(start, end, new):
    """Replaces text on the current line.

    This function replaces from the column number ``start`` to the
    column number ``end`` with the ``new`` text. ``replace`` moves
    the cursor to the starting position and then uses the change
    command (``c``) to replace the text.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - replace: 0, 4, "Snek"

    `Using the API:`

    .. code-block:: python

      ezvi.tools.replace(0, 4, "Snek")

    :type start: int
    :param start: The number of the column to start replacing (inclusively).

    :type end: int
    :param end: The number of the column to stop replacing (exlusively).

    :type new: str
    :param new: The new text to type.

    :rtype: str
    :return: Keystrokes that can be used to replace text from start to end.
    """

    movement = goto_column(start)
    replace = "c" + str(end - start)
    to_write = movement + replace + new + ESCAPE

    return to_write


def find_replace(old, new):
    """To find `old` on the current line and replaces it with `new`."""
    raise NotImplementedError
    pass


def replace_line(new):
    """Replaces text on the current line.

    ``replace`` moves the cursor to the beginning of the line
    using ``0`` from the command mode and then uses command
    (``c$``) to replace the whole line.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - replace_line: "Hello there."

    `Using the API:`

    .. code-block:: python

      ezvi.tools.replace_line("General Kenobi.")

    :type new: str
    :param new: The new text to replace the current line.

    :rtype: str
    :return: Characters that can be typeed to change the current line
    with new content.
    """

    movement = "0"
    replace = "c" + "$"
    to_write = movement + replace + new + ESCAPE

    return to_write


# Vi commands


def write_file(filename):
    """Write the current buffer.

    From the command mode, ``write_file`` uses Vi's ``:w`` command to
    write the current buffer.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - write_file: "message.txt"

    `Using the API:`

    .. code-block:: python

      ezvi.tools.write_file("message.txt")

    :type filename: str
    :param filename: The path towards where to save the current buffer.

    :rtype: str
    :return: ``:w`` plus a file name and a newline.
    """

    to_write = ":w " + filename + "\n"

    return to_write


def quit_editor():
    """To quit the editor.

    From the command mode, ``quit_editor`` uses Vi's ``:q`` command to
    quit the editor. Since Vi makes sure that the current buffer
    is saved before quitting, ``force_quit_editor`` should be used instead
    if the buffer must not be saved.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - quit_editor:

    `Using the API:`

    .. code-block:: python

      ezvi.tools.quit_editor()

    :rtype: str
    :return: ``:q`` and a newline.
    """

    to_write = ":q" + "\n"

    return to_write


def force_quit_editor():
    """To force quit the editor.

    From the command mode, ``force_quit_editor`` uses Vi's ``:q!`` command to
    force quit the editor. All unsaved progress will be lost.

    Usage:

    `In a config file:`

    .. code-block:: yaml

      - force_quit_editor:

    `Using the API:`

    .. code-block:: python

      ezvi.tools.force_quit_editor()

    :rtype: str
    :return: ``:q!`` and a newline character.
    """

    to_write = ":q!" + "\n"

    return to_write


# TODO: Replace ``Any`` with something more accurate.
all_tools: Dict[Any, Callable] = {
    write_chars.__name__: write_chars,
    write_line.__name__: write_line,
    new_line.__name__: new_line,
    new_line_over.__name__: new_line_over,
    write_after_word.__name__: write_after_word,
    write_after_line.__name__: write_after_line,
    write_after_char.__name__: write_after_char,
    write_before_word.__name__: write_before_word,
    write_before_line.__name__: write_before_line,
    write_before_char.__name__: write_before_char,
    goto_line.__name__: goto_line,
    goto_column.__name__: goto_column,
    replace.__name__: replace,
    replace_line.__name__: replace_line,
    write_file.__name__: write_file,
    quit_editor.__name__: quit_editor,
    force_quit_editor.__name__: force_quit_editor,
}



# -*- coding: utf-8 -*-
"""Functions to help with fake typing"""
import os
import time
import random
from typing import List, Dict, Union
from ezvi import tools

LEFT_HAND: List[str] = [
    "as",
    "sa",
    "er",
    "re",
    "sd",
    "ds",
    "ec",
    "ce",
    "ew",
    "we",
    "wa",
    "aw",
    "cr",
    "sc",
    "cs",
]
RIGHT_HAND: List[str] = [
    "lk",
    "lo",
    "ol",
    "op",
    "po",
    "io",
    "oi",
    "no",
    "on",
    "in",
    "ni",
]
HAND_ALTERNATION: List[str] = [
    "al",
    "la",
    "ak",
    "ka",
    "am",
    "ma",
    "an",
    "na",
    "ai",
    "ia",
    "so",
    "os",
    "sp",
    "ps",
    "en",
    "ne",
    "em",
    "me",
    "el",
    "le",
    "ep",
    "pe",
]

PLAUSIBLE_TYPOS: Dict[str, List[str]] = {  # In keyboard order.
    "q": ["w", "a"],
    "w": ["q", "e", "s"],
    "e": ["w", "r", "d"],
    "r": ["e", "t", "f"],
    "t": ["r", "y", "g"],
    "y": ["t", "u", "h"],
    "u": ["y", "i", "j"],
    "i": ["u", "o", "k"],
    "o": ["i", "p", "l"],
    "p": ["o", "[", ";"],
    "a": ["s", "q"],
    "s": ["a", "w", "d"],
    "d": ["s", "f", "e"],
    "f": ["d", "d", "r"],
    "g": ["f", "h", "t"],
    "h": ["g", "j", "y"],
    "j": ["h", "k", "i"],
    "k": ["j", "l", "o"],
    "l": ["k", "o", "p", ";"],
    "z": ["a", "x", "s"],
    "x": ["z", "c", "s"],
    "c": ["x", "v", "d"],
    "v": ["c", "b", "f"],
    "b": ["v", "n", "g"],
    "n": ["b", "m", "h"],
    "m": ["n", "j", "k", ","],
}

ENCODED_BACKSPACE: List[bytes] = tools.ez_encode_str("\b")


def is_typo() -> bool:
    """
    Determines wether or not a combination of key presses should
    generate a typo. For now, we assume that every key press has
    an equal chance of being a typo.

    According to some papers[0], there is between 0.62% and 3.2%
    chances that a keypress will be a typo. For our purposes, we
    will assume that this percentage is between 1 and 3 percent.

    [0]: Refer the documentation.

    Returns:
        bool: Whether or not there will be a typo.

    """
    error_percent = random.randint(1, 4)  # 3.2 rounded up
    # Randint includes the upper bound.
    return random.randint(1, 100) <= error_percent


def is_pause() -> bool:
    """Checks if the computer should pause typing for a while.

    Similar to `is_typo()`, but is not based on any research.
    The chances of taking a small pause has been determined
    by testing different combinations of `is_pause()` and
    `pause_time()`.

    Returns:
        bool: Whether or not the program should stop typing.
    """
    pause_percent: int = random.randint(20, 30)
    return random.randint(1, 100) <= pause_percent


def pause_time() -> float:
    """Returns for how long the program should pause when typing.

    The short pause time is based on what felt more natural when
    testing.

    Returns:
        float: How long the pause shoud last **in seconds**.
    """
    # Between .5 to 1 seconds
    pause_ms: int = random.randint(500, 1000)
    return pause_ms / 1000  # returned pause is in seconds.


def pick_typo(next_letter: str) -> Union[str, None]:
    """Picks a typo according to the next letter to type.

    This function uses `is_typo()` to determine wether or
    not there will be a typo.

    Plausible typos are defined in the `PLAUSIBLE_TYPOS`
    global variable. If `next_letter` is not a key in the
    `PLAUSIBLE_TYPOS` dict, the typo is set to none.

    Args:
        next_letter (str): The next letter that should be typed.

    Returns:
        str/None: The typo or `None` if there is no typo.
    """

    typo: Union[str, None] = None

    if is_typo():

        try:
            plausible_for_letter = PLAUSIBLE_TYPOS[next_letter.lower()]
            # Pick a random typo in the list associated with the
            # key.
            typo = random.choice(plausible_for_letter)

        except KeyError:  # No typo defined for `next_letter`
            typo = None

    else:

        typo = None

    return typo


def get_delay(previous_letter: str, next_letter: str, multiplier: float = 1)  -> float:
    """Function to get the delay before the next letter is typed.

    This function uses the fact that letters typed by two different
    hands will usually be typed 30 to 60ms faster than two letters
    pressed by the same hand.

    There is also an average delay of betweem 120 and 170ms between
    two keystrokes.

    Args:
        previous_letter (str): The previously typed letter.
        next_letter (str): The next letter to type.
        multiplier (int): Typing speed multiplier. The delay is divided
        by the multiplier. Defaults to 1.

    Returns:
        float: The time **in seconds** before the next keystroke.
            This value is calculated using the previous and next
            letter.
    """
    avg_delay = random.randint(120, 170) / 1000  # in seconds
    if (previous_letter + next_letter).lower() in HAND_ALTERNATION:
        # Two letters typed by different hands are 30-60ms faster.
        faster_by = random.randint(30, 60) / 1000
        avg_delay -= faster_by
    return avg_delay / multiplier


def type_typo(file_descriptor: int, next_letter: str, typo: str) -> None:
    """Sends a typo to the file associated with the file descriptor.

    Delays are computed between each keystroke, including when the computer
    presses backspace. For now, only one typo can be introduced at a time.
    More than one would be distracting for the user.

    """
    os.write(file_descriptor, typo.encode("utf-8"))
    time.sleep(get_delay(typo, "backspace"))
    os.write(file_descriptor, "\b".encode("utf-8"))
    time.sleep(get_delay(typo, next_letter))
    os.write(file_descriptor, next_letter.encode("utf-8"))


def type_letter(file_descriptor: int, previous: str, next: str) -> None:
    """Writes the next letter to the file.

    This function also calculates the delay using the previous letter
    and the chances of typing a typo.

    If there is a typo, `type_typo()` is called and writes a typo
    to the file before correcting it.

    If the next character to type is a space, there are chances that the
    program will take a pause. Indeed, people are more likely to hesitate
    between words than while typing a word.
    """
    typo: Union[str, None] = pick_typo(next)

    if next == " ":
        # If the next char to type is a space, compute chances of taking
        # a pause.
        if is_pause():
            time.sleep(pause_time())
    else:
        delay: float = get_delay(previous, next)
        time.sleep(delay)
    if typo:
        type_typo(file_descriptor, next, typo)
    else:
        os.write(file_descriptor, next.encode("utf-8"))



"""
``funcmodule.py`` contains functions that are used to spawn and interact
with a ``vi`` process.
"""
import os
import pty
import time
from typing import Dict, Any, Callable
from pathlib import Path
from ezvi import tools
import yaml
import tty
from select import select
from ezvi import human_typing

STDIN_FILENO = 0
STDOUT_FILENO = 1
CHILD = 0


#######################################################################
#                         Read function                               #
#######################################################################


def ez_read(fd) -> bytes:
    """Standard read function.

    :type fd: int
    :param fd: File descriptor.

    :rtype: bytes
    :return: Up to 1024 bytes that have been read from `fd`.
    """
    return os.read(fd, 1024)


#######################################################################
#                Spawning an writing to process                       #
#######################################################################


def ez_spawn(argv, instructions, master_read=ez_read, stdin_read=ez_read):
    """Spawns a process
    To spawn the process. Heavily inspired from Python's `pty`
    module. `ez_spawn()` should only be used once per file that
    needs to be edited. This function can write a whole text file
    following the instructions dictionary and executing those
    instructions using `ez_copy()`.

    :type argv: tuple
    :param argv: First element is the program to run. Other elements
    are the arguments passed to the program.
    :type instructions: list
    :param instructions: Contains all the instructions that will be
    passed to Vi. Its contents should have been created using the
    "Vi tools".
    :type master_read: function
    :param master_read: The function that will be used to read
    info from the master's file descriptor.
    :type stdin_read: function
    :param stdin_read: The function that will be used to read
    info from the master's file descriptor.

    :rtype: int
    :return: Exit status
    """
    all_written = []  # Useful for debugging.
    if type(argv) == str:
        argv = (argv,)
    pid, master_fd = pty.fork()
    if pid == CHILD:
        # Fork worked, run the program
        os.execlp(argv[0], *argv)
    try:
        mode = tty.tcgetattr(STDIN_FILENO)
        # The next line is required to make sure that you can't type over vi
        # either.
        tty.setraw(STDIN_FILENO)  # disable line buffering
        # interrupt signals are no longer interpreted
        restore = 1
    except tty.error:
        # Did not work, no need to restore.
        restore = 0
    # This is where the fun begins.
    try:
        for item in instructions:
            #  This is where each instruction is written.
            all_written.append(ez_write(master_fd, list(item), master_read))
        time.sleep(.5)
    except OSError:
        if restore:
            # Discard queued data and change mode to original.
            tty.tcsetattr(STDIN_FILENO, tty.TCSAFLUSH, mode)
    if restore:
        tty.tcsetattr(STDIN_FILENO, tty.TCSAFLUSH, mode)
    os.close(master_fd)
    # wait for completion and return exit status
    return os.waitpid(pid, 0)[1]


def ez_write(master_fd, to_write, master_read=ez_read):
    """Writes every char in `to_write` to `master_fd`.

    :type master_fd: int
    :param master_fd: Master's file descriptor.
    :type to_write: list
    :param to_write: List of encoded chars that will be
    written to `master_fd`.
    :type master_read: function
    :param master_read: The function that will be used to read
    info from the master's file descriptor.
    :type stdin_read: function
    :param stdin_read: The function that will be used to read
    from STDIN (if the user wants to write to the program).

    :rtype: list
    :return: A list of all chars that have been written.
    """
    written = []
    fds = [master_fd, STDIN_FILENO]
    while True:
        rfds, wfds, _ = select([fds[0]], [fds[1]], [])
        if master_fd in rfds:
            # This is required to see the program running.
            data = master_read(master_fd)
            if not data:
                fds.remove(master_fd)
            else:
                # Printing the program
                os.write(STDOUT_FILENO, data)
        # This should always be true.
        if STDIN_FILENO in wfds:
            try:
                next_char: str = to_write.pop(0)
            except IndexError:
                next_char = ""
            if not next_char:
                break
            else:
                # Chars a encoded in the type_letter function
                if written: # There is a previous letter.
                    human_typing.type_letter(master_fd, written[-1], next_char)
                else: # This is the first letter to be typed.
                    human_typing.type_letter(master_fd, next_char, next_char)
                written.append(next_char)
    return written


#######################################################################
#                            YAML parsing                             #
#######################################################################


def check_ezvi_config(parsed_config: Any):
    """
    Checks a parsed config file to make sure that it is a valid
    ``ezvi`` configuration file.

    :param parsed_config: The parsed configuration file. Can be
    a Python object of any type, but this function will raise an
    error if it's something other than a ``list``.
    :type parsed_config: Any
    :raises TypeError: If ``parsed_config`` is not of type ``list``.
    :raises TypeError: If an element in ``parsed_config`` is not of
    type ``dict``.
    :raises NotImplementedError: If a command used in the configuration
    file is not part of the available commands.
    """

    if not isinstance(parsed_config, list):
        raise TypeError("A valid ezvi configuration should be parsed as a list.")

    available = [key for key, _ in tools.all_tools.items()]
    for instruction in parsed_config:
        if not isinstance(instruction, dict):
            raise TypeError(
                "An ezvi configuration file should be parsed as a list of dictionaries."
            )
        for key, value in instruction.items():
            # If the function is available, run it with "value"
            # as an argument. This translates the text (values)
            # according to the instructions (values).
            if key in available:
                # Removing None types from dict.
                if instruction[key] is not None:
                    instruction[key] = tools.all_tools[key](value)
                else:
                    instruction[key] = tools.all_tools[key]()
            else:
                raise NotImplementedError(key + " does not exist.")


def yaml_parser(stream) -> list:
    """Loads a YAML file.

    :type stream: textIO
    :param stream: A stream of text to be parsed.

    :rtype: list
    :return: The parsed yaml file.
    """

    parsed = yaml.safe_load(stream)
    check_ezvi_config(parsed)

    return parsed


#######################################################################
#                       File comparison                               #
#######################################################################

# To convert an existing text file to some ezvi commands.


def path_check(name):
    """To check if a path to save is valid.

    :param name: The name of the file.
    :type name: str

    :raises Exception: If the path is not valid.

    :return: True if the path is valid.
    :rtype: bool
    """
    if Path(name).is_file():
        raise Exception("{} already exists.".format(name))
    else:
        return True


def file_parser(stream, name=""):
    """To parse a pre-typed text file.

    :param stream: The stream of text to parse.
    :type stream: textIO

    :param save: Where the file should be saved.
    :type save: str

    :return: The parsed text.
    :rtype: list
    """

    to_return = []

    file = stream.readlines()
    for line in file:
        to_return.append(tools.write_chars(line))

    if name == "":
        to_return.append(tools.force_quit_editor())
    else:
        if path_check(name):
            to_return.append(tools.write_file(name))
            to_return.append(tools.quit_editor())

    return to_return


def new_conf(stream, savepath):
    """To create a new configuration file.

    :param stream: The stream of text to create a config from.
    :type stream: textIO
    :param savepath: The path to save the new config file.
    :type savepath: str
    :return: None
    :rtype: NoneType
    """
    to_write = []

    if path_check(savepath):
        file = open(savepath, "w")

    for line in stream:
        line = line.strip()
        if not line:
            to_append = "- new_line: "
        else:
            to_append = "- write_line: " + line + "\n"
        to_write.append(to_append)

    file.writelines(to_write)


#######################################################################
#                      Searching/editing tools                        #
#######################################################################



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ezvi']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'astroid==2.4.2',
 'click==7.1.2',
 'isort==5.6.4',
 'lazy-object-proxy==1.4.3',
 'mccabe==0.6.1',
 'pylint==2.6.0',
 'six==1.15.0',
 'toml==0.10.2']

entry_points = \
{'console_scripts': ['ezvi = ezvi.cli:main']}

setup_kwargs = {
    'name': 'ezvi',
    'version': '0.1.7',
    'description': 'Automated typing in the Vi editor.',
    'long_description': '# ezvi\n\n`ezvi` is a python package that allows typing automation for the `Vi` editor.\n\n## Installation\n\n`ezvi` is distributed as a Pip package. To install the program, simply run\n\n```bash\npip install ezvi\n```\n\n## Usage\n\nThe package can be used via the `CLI`. `ezvi` functions can also be imported and used in a Python program.\n\n### The `CLI`\n\nThere are two different ways of using `ezvi` via the command line. \n\n#### The `text` command \n\n`text` can be used to type a pre-written file. It takes one argument (`infile` and one option (`--writefile`).\n\n* `infile` is the path towards the pre-written file.\n* `--writefile` tells the program to save the file again after typing it. `--writefile` takes one argument that corresponds to the path where the typed file will be written.\n\n##### Example\n\n```bash\nezvi text -w ./foo.txt example/message.txt\n```\n\nThis takes the `message.txt` example from the [example](https://github.com/TrickyTroll/ezvi/tree/latest/example) directory and types it again. The Vi buffer is then written to `./foo.txt`.\n\n#### The `yaml` command\n\n`yaml` should be used to take a configuration as instructions to type a new file. The `yaml` command only takes one argument (`config`). Everything else should be specified in the config file.\n\n* `config` is the path towards the configuration file.\n\n##### Example\n\n```bash\nezvi yaml example/config.yaml\n```\n\nThis command would take the `config.yaml` file from the [example](https://github.com/TrickyTroll/ezvi/tree/latest/example) directory and use it to type a new file.\n\n### Writing a config file\n\nA configuration file is just a yaml file that will be parsed using [PyYAML](https://pyyaml.org "PyYAML"). The structure of the file should be similar to the one in the `config.yaml` file from the [example](https://github.com/TrickyTroll/ezvi/tree/latest/example) directory.\n\n```yaml\n- write_line: "Hello!"\n- new_line: 2\n- write_chars: "-- Good Bot."\n- write_file: "message.txt"\n- quit_editor:\n```\n\nA `-` must precede every action.\n\n## Documentation\n\nFor more in depth documentation, please refer to the \n[Read the Docs](https://ezvi.readthedocs.io/en/main/)\n\n## Development\n\nThis package is in alpha. Not much testing has been done and many things could still change.  To see the latest commit, go check the [latest](https://github.com/TrickyTroll/ezvi/tree/latest) branch.\n',
    'author': 'TrickyTroll',
    'author_email': 'tricky@beon.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
